clc
clear all;
%% Parameters
h = 2;
Dt = 0.1;
a_max = 9.8;
a_min = 9.8;

IsVerbose = 1;

% Assumptions on leader car:
% - The leader car follows the laws of kinematics
%       p_2^+ = p_2 + Dt*v_2
%       v_2^+ = v_2 + Dt*a_2
% - a2 must have bounds:
%       -a_min <= a2 <= a_max
%
% Guarantees:
% - Headway is preserved:
%       p_2 - p_1 - h*v_1 >= 0
%
%
% System:
% - The follower car follows the laws of kinematics
%       p_1^+ = p_1 + Dt*v_1
%       v_1^+ = v_1 + Dt*a_1
% - a1 is a linear controller
%       a_1 = (p_2 - p_1) / (h*Dt) - (1/h + 1/Dt)*v_1 + v_2/h

%% Variables
sdpvar p1 p2 v1 v2 a1 a2 p1_plus p2_plus v1_plus v2_plus

%% Constraints
Constraints = [];

%Guarantees of previous step
Constraints = [Constraints,p2 - p1 - h*v1 >= 0];

%Assumptions on input - other car
Constraints = [Constraints,p2_plus == p2 + Dt*v2];
Constraints = [Constraints,v2_plus == v2 + Dt*a2];
Constraints = [Constraints,a2 <= a_max];
Constraints = [Constraints,a2 >= -a_min];

%Assumptions on input - follower acceleration
Constraints = [Constraints, a1 == 0.9*(p2-p1-h*v1)/(h*Dt) - (v1-v2)/h];

%System
Constraints = [Constraints, p1_plus == p1 + Dt*v1];
Constraints = [Constraints, v1_plus == v1 + Dt*a1];

%% Objectives

% We want that the minimum of the objective function is >= Threshold.

Objective(1) = p2_plus - p1_plus - h*v1_plus;

Threshold(1) = 0;


%% Solve
tic;
options = sdpsettings('verbose',IsVerbose);
Val = zeros(size(Objective));
toc
for i=1:length(Objective)
    sol = optimize(Constraints,Objective(i),options);
    Val(i) = value(Objective(i));
    if(sol.problem ~= 0)
        Val(i) = nan;
    end
    if(IsVerbose)
        if(sol.problem == 0) 
            disp(['Optimization problem #',num2str(i),' verifies guarantee only if:']);
            disp([num2str(Val(i)),' >= ', num2str(Threshold(i))]);
        else
            disp(['Optimization problem #',num2str(i),' could not be solved (unbounded, infeasible, or for another reason)']);
        end
    end
end

disp('*********');
PassingGrades = (Val >= Threshold);
if(sum(~PassingGrades) == 0)
    disp('Contract is Verified');
else
    disp('Contract is not verified - Here is why:');
    for j=1:length(Val)
        if(PassingGrades(j) == 0)
            if(isnan(Val(j)))
            disp(['Optimization problem #',num2str(j),' could not be solved (unbounded, infeasible, or for another reason)']);
            else
            disp(['Guarantee #',num2str(j),' cannot be assured, as the value of']);
            disp(['the optimization problem is ',num2str(Val(j)),', and the']);
            disp(['Thershold, which is supposed to be a lower bound, is ',num2str(Threshold(j)),'.']);
            disp(' ');
            end
        end
    end
end